#!/usr/bin/env python3
"""
Setup script for torchada.

This is a shim for editable installs. The actual configuration is in pyproject.toml.
"""

from setuptools import setup

if __name__ == "__main__":
    setup()

